/************************************************************************
Copyright Notice

Copyright (c) MCNC, Clearinghouse for Networked Information Discovery and
Retrieval, 1994. 

Permission to use, copy, modify, distribute, and sell this software and
its documentation, in whole or in part, for any purpose is hereby granted
without fee, provided that

1. The above copyright notice and this permission notice appear in all
copies of the software and related documentation. Notices of copyright
and/or attribution which appear at the beginning of any file included in
this distribution must remain intact. 

2. Users of this software agree to make their best efforts (a) to return
to MCNC any improvements or extensions that they make, so that these may
be included in future releases; and (b) to inform MCNC/CNIDR of noteworthy
uses of this software. 

3. The names of MCNC and Clearinghouse for Networked Information Discovery
and Retrieval may not be used in any advertising or publicity relating to
the software without the specific, prior written permission of MCNC/CNIDR. 

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY
OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. 

IN NO EVENT SHALL MCNC/CNIDR BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR
IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
************************************************************************/

/*@@@
File:		Iindex.cxx
Version:	1.00
Description:	Command-line indexer
Author:		Nassib Nassar, nrn@cnidr.org
@@@*/

#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include "db.hxx"
#include "string.hxx"
#include "common.hxx"
#include "record.hxx"
#include "reclist.hxx"

#include <iostream.h>

class IDB : public DB {
public:
	IDB(const STRING& NewPathName, const STRING& NewFileName) :
			DB(NewPathName, NewFileName) { };
protected:
	void IndexingStatus(const INT StatusMessage, const PSTRING FileName,
			const INT WordCount) const {
		switch (StatusMessage) {
			case (IndexingStatusReading):
				printf("   Reading file ");
				FileName->Print();
				printf(" ...\n");
				break;
			case (IndexingStatusIndexing):
				printf("   Indexing %d words ...\n",
						WordCount);
				break;
			case (IndexingStatusMerging):
				printf("   Merging index ...\n");
				break;
		}
	};
};

typedef IDB* PIDB;

STRING Separator;

void AddFile(PRECLIST Reclist, STRING& PathName, STRING& FileName) {
	RECORD Record;
	Record.SetPathName(PathName);
	Record.SetFileName(FileName);
	if (Separator.Equals("")) {
		Record.SetRecordStart(0);
		Record.SetRecordEnd(0);
		Reclist->AddEntry(Record);
	} else {
		GPTYPE Start = 0;
		GPTYPE Position = 0;
		GPTYPE NewPosition;
		STRING Fn;
		CHR Ch, Sch;
		Sch = Separator.GetChr(1);
		STRINGINDEX Slen = Separator.GetLength();
		CHR Buffer[Slen+1];
		Record.GetFullFileName(&Fn);
		PFILE Fp = fopen(Fn, "r");
		while (fread(&Ch, 1, 1, Fp) == 1) {
			Position++;
			if ( (Ch == Sch) && (Position > (Start+1)) ) {
				fseek(Fp, --Position, 0);
				NewPosition = Position +
						fread(&Buffer, 1, Slen, Fp);
				Buffer[Slen] = '\0';
				if (Separator.Equals(Buffer)) {
					Record.SetRecordStart(Start);
					Record.SetRecordEnd(Position - 1);
					Reclist->AddEntry(Record);
					Start = Position;
				}
				Position = NewPosition;
			}
		}
		if ( Position > (Start+1) ) {
			Record.SetRecordStart(Start);
			Record.SetRecordEnd(Position);
			Reclist->AddEntry(Record);

		}
		fclose(Fp);
	}
}

int main(int argc, char** argv) {
	if (argc < 2) {
		printf("\nIindex, Version 1.00, Copyright (c) 1994 MCNC/CNIDR\n\n");
		printf("Iindex [-d (X)]  // Use (X) as the base name for index files.\n");
		printf("       [-a]  // Add to existing database, instead of replacing it.\n");
		printf("       [-m (X)]  // Use (X) megabytes of memory for indexing (default=2).\n");
		printf("       [-s (X)]  // Treat (X) as a separator for multiple records within a\n");
		printf("                 // single file.\n");
		printf("       [-f (X)]  // Read list of file names to be indexed from file (X).\n");
		printf("       [-r]  // Recursively descend subdirectories.\n");
		printf("       (X) (Y) (...)  // Index files (X), (Y), etc.\n\n");
		printf("Example:  Iindex -d BLAKE *.doc *.txt\n\n");
		return 0;
	}
	CHR Cwd[256];
	getcwd(Cwd, 255);
	STRING Flag;
	STRING DBName;
	STRING FileList;
	INT Recursive = 0;
	INT AppendDb = 0;
	UINT4 MemoryUsage = 0;
	INT x = 0;
	INT LastUsed = 0;
	while (x < argc) {
		if (argv[x][0] == '-') {
			Flag = argv[x];
			if (Flag.Equals("-d")) {
				if (++x >= argc) {
					printf("ERROR: No database name specified after -d.\n\n");
					return 0;
				}
				DBName = argv[x];
				LastUsed = x;
			}
			if (Flag.Equals("-f")) {
				if (++x >= argc) {
					printf("ERROR: No file name specified after -f.\n\n");
					return 0;
				}
				FileList = argv[x];
				LastUsed = x;
			}
			if (Flag.Equals("-s")) {
				if (++x >= argc) {
					printf("ERROR: No separator string specified after -s.\n\n");
					return 0;
				}
				Separator = argv[x];
				LastUsed = x;
			}
			if (Flag.Equals("-m")) {
				if (++x >= argc) {
					printf("ERROR: No memory usage specified after -m.\n\n");
					return 0;
				}
				MemoryUsage = strtol(argv[x], NULL, 10);
				LastUsed = x;
			}
			if (Flag.Equals("-a")) {
				AppendDb = 1;
				LastUsed = x;
			}
			if (Flag.Equals("-r")) {
				Recursive = 1;
				LastUsed = x;
			}
		}
		x++;
	}

	if (DBName.Equals("")) {
		printf("ERROR: No database name specified!\n");
		return 0;
	}

	x = LastUsed + 1;

	INT NumFiles = argc - x;
	INT z = x;

	if ((FileList.Equals("")) && (NumFiles == 0)) {
		printf("ERROR: No files specified for indexing!\n");
		return 0;
	}

	if ( (!FileList.Equals("")) && (NumFiles != 0) ) {
		printf("ERROR: Unable to handle -f and file names at the same time.\n");
		return 0;
	}

	STRING IndexFN, MdtFN, TempFN;	// for filtering - should be done in library!
	if (!AppendDb) {
		IndexFN = DBName;
		IndexFN.Cat(".inx");
		MdtFN = DBName;
		MdtFN.Cat(".mdt");
		TempFN = DBName;
		TempFN.Cat(".tmp");
		unlink(IndexFN);
		unlink(MdtFN);
		unlink(TempFN);
	}

	RECLIST reclist;
//	RECORD record;
	STRING PathName, FileName;

	printf("Building record list ...\n");

	if ( FileList.Equals("") ) {
		INT y;
		for (z=0; z<NumFiles; z++) {
			if ( (!IndexFN.Equals(argv[z+x])) &&
					(!MdtFN.Equals(argv[z+x])) &&
					(strcmp(argv[z+x], "core") != 0) ) {
				y = chdir(argv[z+x]);
				if (y == -1) {
					PathName = argv[z+x];
					FileName = argv[z+x];
					RemovePath(&FileName);
					RemoveFileName(&PathName);
					AddFile(&reclist, PathName, FileName);
				} else {
					chdir(Cwd);
					if (Recursive) {
						CHR tempbuf[256];
						sprintf(tempbuf, "find %s -name \"*\" -print > Iindex.tmp",
								argv[z+x]);
//printf("system: [%s]\n", tempbuf);
						system(tempbuf);
//gets(tempbuf);

						
						PFILE fp = fopen("Iindex.tmp", "r");
						CHR s[256], t[256];
						STRING v;
						if (!fp) {
							printf("ERROR: Can't generate file list (-r).\n");
							return 0;
						}
						while (fgets(s, 255, fp) != NULL) {
							if (s[strlen(s)-1] == '\n') {
								s[strlen(s)-1] = '\0';
							}
							v = s;
							RemovePath(&v);
							v.GetCString(t, 256);
							if ( (!IndexFN.Equals(t)) &&
									(!MdtFN.Equals(t)) &&
									(strcmp(t, "Iindex.tmp") != 0) &&
									(strcmp(t, "core") != 0) &&
									(strlen(t) > 0) ) {
								y = chdir(s);
								if (y == -1) {
									PathName = s;
									FileName = s;
									RemovePath(&FileName);
									RemoveFileName(&PathName);
									AddFile(&reclist, PathName, FileName);
								} else {
									chdir(Cwd);
								}
							}
						}
						fclose(fp);
						unlink("Iindex.tmp");
					}
				}
			}
		}
	}

	if ( !FileList.Equals("") ) {
		PFILE fp = fopen(FileList, "r");
		CHR s[256];
		if (!fp) {
			printf("ERROR: Can't find file list (-f).\n");
			return 0;
		}
		while (fgets(s, 255, fp) != NULL) {
			if (s[strlen(s)-1] == '\n') {
				s[strlen(s)-1] = '\0';
			}
			PathName = s;
			FileName = s;
			RemovePath(&FileName);
			RemoveFileName(&PathName);
			AddFile(&reclist, PathName, FileName);
		}
		fclose(fp);
	}

	PIDB pdb;
	STRING DBPathName, DBFileName;

	DBPathName = DBName;
	DBFileName = DBName;
	RemovePath(&DBFileName);
	RemoveFileName(&DBPathName);
	pdb = new IDB(DBPathName, DBFileName);

	if (AppendDb) {
		printf("Adding to database ");
	} else {
		printf("Building database ");
	}
	DBName.Print();
	printf(":\n");

	if (MemoryUsage != 0) {
		pdb->SetIndexingMemory(MemoryUsage*1024*1024);
	}
	printf("Using %i MB of system memory.\n",
			pdb->GetIndexingMemory() / 1024 / 1024);

	pdb->AddRecordList(reclist);

	delete pdb;
	printf("Database files saved to disk.\n");

	unlink("tmp");
	return 0;
}
