/************************************************************************
Copyright Notice

Copyright (c) MCNC, Clearinghouse for Networked Information Discovery and
Retrieval, 1994. 

Permission to use, copy, modify, distribute, and sell this software and
its documentation, in whole or in part, for any purpose is hereby granted
without fee, provided that

1. The above copyright notice and this permission notice appear in all
copies of the software and related documentation. Notices of copyright
and/or attribution which appear at the beginning of any file included in
this distribution must remain intact. 

2. Users of this software agree to make their best efforts (a) to return
to MCNC any improvements or extensions that they make, so that these may
be included in future releases; and (b) to inform MCNC/CNIDR of noteworthy
uses of this software. 

3. The names of MCNC and Clearinghouse for Networked Information Discovery
and Retrieval may not be used in any advertising or publicity relating to
the software without the specific, prior written permission of MCNC/CNIDR. 

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY
OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. 

IN NO EVENT SHALL MCNC/CNIDR BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR
IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
************************************************************************/

/*@@@
File:		irset.cxx
Version:	1.00
Description:	Class IRSET - Internal Search Result Set
Author:		Nassib Nassar, nrn@cnidr.org
@@@*/

#include <stdlib.h>
#include "irset.hxx"
#include "common.hxx"

IRSET::IRSET(const PDBOBJ DbParent) {
	Table = new IRESULT[100];
	TotalEntries = 0;
	MaxEntries = 100;
	Parent = DbParent;
}

void IRSET::LoadTable(const STRING& FileName) {
	PFILE fp = fopen(FileName, "r");
	if (fp) {
		INT FSize = GetFileSize(fp);
		INT TSize = sizeof(IRESULT);
		INT ArraySize = (FSize / TSize) + 1;
		Resize(ArraySize);
		TotalEntries = fread(Table, 1, FSize, fp) / TSize;
		fclose(fp);
		MaxEntries = ArraySize;
	}
}

void IRSET::SaveTable(const STRING& FileName) {
	PFILE fp = fopen(FileName, "w");
	if (fp) {
		fwrite(Table, 1, sizeof(IRESULT)*TotalEntries, fp);
		fclose(fp);
	}
}

void IRSET::AddEntry(const IRESULT& ResultRecord, const INT AddHitCounts) {
	INT x;
	// linear!
	for (x=0; x<TotalEntries; x++) {
		if (Table[x].GetMdtIndex() == ResultRecord.GetMdtIndex()) {
			if (AddHitCounts) {
				Table[x].IncHitCount(
						ResultRecord.GetHitCount());
			}
			Table[x].IncScore(ResultRecord.GetScore());
			return;
		}
	}

	if (TotalEntries == MaxEntries)
		Expand();
	Table[TotalEntries] = ResultRecord;
	TotalEntries = TotalEntries + 1;
}

void IRSET::GetEntry(const INT Index, PIRESULT ResultRecord) const {
	if ( (Index > 0) && (Index <= TotalEntries) ) {
		*ResultRecord = Table[Index-1];
	}
}

PRSET IRSET::GetRset() {
	PRSET prset = new RSET;
	RESULT result;
	MDTREC mdtrec;
	STRING s;
	INT x;
	for (x=0; x<TotalEntries; x++) {
		Parent->GetMainMdt()->GetEntry(Table[x].GetMdtIndex(), &mdtrec);
		mdtrec.GetKey(&s);
		result.SetKey(s);
		mdtrec.GetPathName(&s);
		result.SetPathName(s);
		mdtrec.GetFileName(&s);
		result.SetFileName(s);
		result.SetRecordStart(mdtrec.GetLocalRecordStart());
		result.SetRecordEnd(mdtrec.GetLocalRecordEnd());
		result.SetScore(Table[x].GetScore());
		prset->AddEntry(result);
	}
	return prset;
}

void IRSET::Expand() {
	Resize(TotalEntries+100);
}

void IRSET::CleanUp() {
	Resize(TotalEntries);
}

void IRSET::Resize(const INT Entries) {
	PIRESULT Temp = new IRESULT[Entries];
	INT RecsToCopy;
	INT x;
	if (Entries >= TotalEntries) {
		RecsToCopy = TotalEntries;
	} else {
		RecsToCopy = Entries;
		TotalEntries = Entries;
	}
	for (x=0; x<RecsToCopy; x++) {
		// Not sure if Temp[x] = Table[x] is good enough.
		Temp[x] = Table[x];
	}
	if (Table)
		delete [] Table;
	Table = Temp;
	MaxEntries = Entries;
}

INT IRSET::GetTotalEntries() const {
	return TotalEntries;
}

INT IRSET::GetHitTotal() const {
	INT x;
	INT Total = 0;
	for (x=0; x<TotalEntries; x++) {
		Total += Table[x].GetHitCount();
	}
	return Total;
}

void IRSET::Or(const IRSET& OtherIrset) {
	INT x;
	IRESULT OtherIresult;
	for (x=1; x<=OtherIrset.GetTotalEntries(); x++) {
		OtherIrset.GetEntry(x, &OtherIresult);
		AddEntry(OtherIresult, 0);
	}
}

void IRSET::ComputeScores() {
	INT x;
	DOUBLE DocsInRs = TotalEntries;
	DOUBLE DocsInDb = Parent->GetMainMdt()->GetTotalEntries();
	DOUBLE InvDocFreq = DocsInDb / DocsInRs;
	DOUBLE SumSqScores = 0;
	DOUBLE SqrtSum;
	DOUBLE Score;
	for (x=0; x<TotalEntries; x++) {
		Score = Table[x].GetHitCount() * InvDocFreq;
		Table[x].SetScore(Score);
		SumSqScores += (Score * Score);
	}
	SqrtSum = sqrt(SumSqScores);
	if (SqrtSum == 0.0) {
		SqrtSum = 1.0;
	}
	for (x=0; x<TotalEntries; x++) {
		Table[x].SetScore(Table[x].GetScore() / SqrtSum);
	}
}

int IrsetScoreCompare(const void* x, const void* y) {
	return ( (INT)( ((*((PIRESULT)y)).GetScore())*1000 -
			((*((PIRESULT)x)).GetScore())*1000 ) );
}

void IRSET::SortByScore() {
	qsort(Table, TotalEntries, sizeof(IRESULT), IrsetScoreCompare);
}

IRSET::~IRSET() {
	if (Table)
		delete [] Table;
}
