#define CALLOC(a,b) calloc(a,b)
#define FREE(a) free(a)
#include "imem.h"

/*@@@
File:           netbox.c
Version:        1.01.01
Description:    Socket-interfacing network functions
Author:         Nassib Nassar; Kevin Gamiel
@@@*/

/*************************************************************************

Copyright Notice

Copyright (c) MCNC, Clearinghouse for Networked Information Discovery and
Retrieval, 1994. 

Permission to use, copy, modify, distribute, and sell this software and
its documentation, in whole or in part, for any purpose is hereby granted
without fee, provided that

1. The above copyright notice and this permission notice appear in all
copies of the software and related documentation. Notices of copyright
and/or attribution which appear at the beginning of any file included in
this distribution must remain intact. 

2. Users of this software agree to make their best efforts (a) to return
to MCNC any improvements or extensions that they make, so that these may
be included in future releases; and (b) to inform MCNC/CNIDR of noteworthy
uses of this software. 

3. The names of MCNC and Clearinghouse for Networked Information Discovery
and Retrieval may not be used in any advertising or publicity relating to
the software without the specific, prior written permission of MCNC/CNIDR. 

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY
OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. 

IN NO EVENT SHALL MCNC/CNIDR BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR
IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 

*************************************************************************/

#include <stdio.h>
#include <string.h>
#include <errno.h>

#ifdef PLATFORM_WINDOWS
#include <sys\types.h>
#else
#include <sys/types.h>
#include <arpa/nameser.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/time.h>
#include <resolv.h>
#endif
#include <malloc.h>
#include "netbox.h"

/*
Returns pointer to NETBOXPROFILE structure allocated using imem on success.
Returns NULL on failure.
*/
PNETBOXPROFILE netbox_Create()
{
	return (PNETBOXPROFILE)CALLOC(1,sizeof(NETBOXPROFILE));
}

void netbox_Destroy(PNETBOXPROFILE P)
{
	FREE(P);
}

char FAR *netbox_SetHostName(PNETBOXPROFILE P, char FAR *Name)
{
	if(strlen(Name)>255)
		return NULL;

	strcpy(P->HostName, Name);
	
	return P->HostName;
}

int netbox_SetPort(PNETBOXPROFILE P, int Port)
{
	if(Port<0)
		return -1;

	P->Port = Port;
	return P->Port;
}

long netbox_SetTimeOutSec(PNETBOXPROFILE P, long T)
{
	if(T<0)
		return -1;
	P->TimeOutSec = T;

	return P->TimeOutSec;
}

long netbox_SetTimeOutUSec(PNETBOXPROFILE P, long T)
{
	if(T<0)
		return -1;
	P->TimeOutUSec = T;

	return P->TimeOutUSec;
}

int netbox_GetPort(PNETBOXPROFILE P)
{
	return P->Port;
}

/*
Returns a pointer to the host name.
*/
char FAR *netbox_GetHostName(PNETBOXPROFILE P)
{
	return P->HostName;
}

int netbox_WSAStartup(void) {
#ifdef PLATFORM_WINDOWS
	WSADATA WsaData;
	if ( WSAStartup(0x0101, &WsaData) != 0 )
		return -1;
#endif
	return 0;
}

void netbox_WSACleanup(void) {
#ifdef PLATFORM_WINDOWS
	WSACleanup();
#endif
}

int netbox_SetBufferSize(PNETBOXPROFILE profile, int BufferSize) {
	int BufSize = BufferSize;
	return setsockopt(profile->Socket, SOL_SOCKET, SO_RCVBUF,
			(char FAR*)&BufSize, sizeof(BufSize));
}

int netbox_Open(PNETBOXPROFILE profile) {
	SOCKET s;
	char HostNameBuf[256];
	LPHOSTENT HostEnt;
	SOCKADDR_IN addr_in;
	int rc;

	memset((char*)&addr_in, 0, sizeof(addr_in));
	addr_in.sin_addr.s_addr = inet_addr(profile->HostName);
	if (addr_in.sin_addr.s_addr != -1) {
		addr_in.sin_family = AF_INET;
		strcpy(HostNameBuf, profile->HostName);
	} else {
		HostEnt = gethostbyname(profile->HostName);
		if ( NULL == HostEnt ) {
			return 0;
		}
		addr_in.sin_family = HostEnt->h_addrtype;
#ifdef h_addr
		strncpy((char*)&addr_in.sin_addr, HostEnt->h_addr_list[0],
				HostEnt->h_length);
#endif
		strcpy(HostNameBuf, HostEnt->h_name);
	}
	addr_in.sin_port = htons(profile->Port);
	if ( (s=socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET )
		return 0;
	rc = connect(s, (struct sockaddr FAR*)&addr_in, sizeof(addr_in));
	if (rc == -1)
		return 0;
	profile->Socket = s;
	return 1;
}

int netbox_Close(PNETBOXPROFILE profile) {
	if ( closesocket(profile->Socket) == 0 ) {
		return 1;
	} else {
		return 0;
	}
}

int netbox_SendBuffer(PNETBOXPROFILE profile, char FAR* Buffer,
		int BufferSize) {

	if((!profile) || (!Buffer) || (BufferSize <= 0))
		return -2;
	/* select() checking should be inserted here */
#ifdef PLATFORM_WINDOWS
	if ( send(profile->Socket, Buffer, BufferSize, 0) == SOCKET_ERROR )
		return -1;
#else
/*
	if(profile->ServerOwner != NB_COMMANDLINE)
		fwrite(Buffer, 1, BufferSize, stdout);
	else	
*/
		if ( write(profile->Socket, Buffer, BufferSize) == -1 )
			return -1;
#endif
	return BufferSize;
}

int netbox_iGetBuffer(PNETBOXPROFILE profile, char FAR* Buffer,
		int BufferSize) {
	int x;
	int TotalRead = 0;
	int Left;
	char* p;
	p = Buffer;
	Left = BufferSize;
	do {
#ifdef PLATFORM_WINDOWS
		x = recv(profile->Socket, p, Left, 0);
#else
		if(netbox_WaitForData(profile)==0)
			return -2;
		x = read(profile->Socket, p, Left);
#endif
		if ( (x != SOCKET_ERROR) && (x != 0) ) {
			p += x;
			Left -= x;
			TotalRead += x;
		}
	} while ( (x != SOCKET_ERROR) && (x != 0) &&
			(TotalRead < BufferSize) );
	if (x == SOCKET_ERROR)
		return -1;
	else
		return TotalRead;
}

long netbox_GetBuffer(PNETBOXPROFILE profile, char FAR* Buffer,
		long BufferSize) {
	int x = 1;
	int y;
	char* p = Buffer;
	long Left = BufferSize;
	long TotalRead = 0;
	while ( (x != 0) && (TotalRead < BufferSize) ) {
		if (Left > MAXINT)
			y = MAXINT;
		else
			y = (int)Left;
		x = netbox_iGetBuffer(profile, p, y);
		if (x == -1)
			return -1;
		if (x != 0) {
			TotalRead += x;
			Left -= x;
			p += x;
		}
	}
	return TotalRead;
}

int netbox_WaitForData(PNETBOXPROFILE profile) {
	fd_set ReadFDS;
	struct timeval TimeVal;
	TimeVal.tv_sec = profile->TimeOutSec;
	TimeVal.tv_usec = profile->TimeOutUSec;
	FD_ZERO(&ReadFDS);
	FD_SET(profile->Socket, &ReadFDS);
	/* Check the socket for readability. */
	if ( (select((profile->Socket)+1, &ReadFDS, 0, 0, &TimeVal)) <= 0 ) {
		/* Timed out or SOCKET_ERROR! Not ready to write. */
		return 0;
	}
	return 1;
}

int netbox_StartServer(PNETBOXPROFILE Profile) {
	SOCKET s;
	SOCKADDR_IN addr_in;

	if(Profile->ServerOwner == 0) {
		/* Server already started by INETD */
		return(0);
	}

	memset(&addr_in, 0, sizeof(addr_in));
	if ( (s=socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET )
		return -1;
	addr_in.sin_family = AF_INET;
	addr_in.sin_addr.s_addr = htonl(INADDR_ANY);
	addr_in.sin_port = (unsigned short int)htons(Profile->Port);
	if ( bind(s, (struct sockaddr*)&addr_in, sizeof(addr_in)) ==
			SOCKET_ERROR ) {
#ifdef PLATFORM_WINDOWS
		if (WSAGetLastError() == WSAEADDRINUSE) {
			if ( connect(s, (const struct sockaddr FAR*)&addr_in,
					sizeof(addr_in)) == 0 ) {
				netbox_WSACleanup();
#else
		if (errno == EADDRINUSE) {
			if ( connect(s, (struct sockaddr*)&addr_in,
					sizeof(addr_in)) == 0 ) {
#endif
				close(s);
				return -1;
			}
		} else {
			int test = 1;
#ifdef PLATFORM_WINDOWS
			netbox_WSACleanup();
#endif
			close(s);
			if ( (s=socket(AF_INET, SOCK_STREAM, 0)) ==
					INVALID_SOCKET )
				return -1;
#ifdef SOLARIS
			setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&test,
					sizeof(test));
#else
			setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &test,
					sizeof(test));
#endif
			addr_in.sin_family = AF_INET;
			addr_in.sin_addr.s_addr = INADDR_ANY;
			addr_in.sin_port = (unsigned short int)htons(Profile->Port);
			if ( bind(s, (struct sockaddr*)&addr_in, sizeof(addr_in)) ==
				SOCKET_ERROR )
				return -1;
		}
	}
	if ( listen(s, 8) == -1 ) {
#ifdef PLATFORM_WINDOWS
		netbox_WSACleanup();
#endif
		close(s);
		return -1;
	}
	Profile->Socket = s;
	return(0);
}

PNETBOXPROFILE netbox_AcceptClient(PNETBOXPROFILE Profile) {
	SOCKADDR_IN addr_in;
	int addrSize;
	PNETBOXPROFILE NewSocket;

	NewSocket = (PNETBOXPROFILE)imem_calloc(1,sizeof(NETBOXPROFILE));
	if(!NewSocket)
		return NULL;

	strcpy(NewSocket->HostName, Profile->HostName);
	NewSocket->Port = Profile->Port;
	NewSocket->ServerOwner = Profile->ServerOwner;
	NewSocket->TimeOutSec = Profile->TimeOutSec;
	NewSocket->TimeOutUSec = Profile->TimeOutUSec;

	if(Profile->ServerOwner == 0) {
		/* Server already started by INETD */
		return(0);
	}

	addrSize = sizeof(addr_in);
 
	do {
#ifdef NETBOX_SOCKADDR_CAST
		if ( (NewSocket->Socket = accept(Profile->Socket,
			(struct sockaddr *)&addr_in, 
			&addrSize)) ==
#else
		if ( (NewSocket->Socket = accept(Profile->Socket,&addr_in, 
			&addrSize)) ==
#endif
				INVALID_SOCKET) {
			switch(errno) {
				case EBADF:
					printf("ebadf\n");
					break;
				case ENOTSOCK:
					printf("enotsock\n");
					break;
				case EOPNOTSUPP:
					printf("eopnotsupp\n");
					break;
				case EFAULT:
					printf("efault\n");
					break;
				case EWOULDBLOCK:
					printf("ewouldblock\n");
					break;
				default:
					printf("dunno\n");

			}
			return NULL;
		}
#ifdef PLATFORM_WINDOWS
	} while (WSAGetLastError() == WSAEINTR);
#else
	} while (errno == EINTR);
#endif
	netbox_CurrentStatus(Profile);
	netbox_CurrentStatus(NewSocket);
	return NewSocket;		
}

int netbox_CurrentStatus(PNETBOXPROFILE Profile) {
        SOCKADDR_IN Name;
        int NameLen;
        struct hostent *Peer;

	NameLen = sizeof(SOCKADDR_IN);
#ifdef NETBOX_SOCKADDR_CAST
	if(!getpeername(fileno(stdout),(struct sockaddr *)&Name,&NameLen)) {
#else
	if(!getpeername(fileno(stdout),&Name,&NameLen)) {
#endif
       		Profile->ServerOwner = 0; 
	}else
       		Profile->ServerOwner = 1; 
        
#ifdef NETBOX_SOCKADDR_CAST
	getpeername(Profile->Socket,(struct sockaddr *)&Name,&NameLen);
#else
	getpeername(Profile->Socket,&Name,&NameLen);
#endif
	Peer = gethostbyaddr(inet_ntoa(Name.sin_addr), 4, AF_INET);
        if(Peer != NULL) {
		sprintf(Profile->PeerName,"%s [%s]",Peer ? Peer->h_name:"?",
			inet_ntoa(Name.sin_addr));
        } else {
		sprintf(Profile->PeerName,"[%s]", inet_ntoa(Name.sin_addr));
        }
	return(0);
}

/* 
PRE:	NetProfile is connected.
	0 < sizeof(Buffer) < MaxLen
POST:	Buffer contains a NULL-terminated string from the net on success.
	On failure, returns <= 0.  See netbox_iGetBuffer for error
	codes.
NOTE:	Copies a single character from the net into Buffer until either
	CR or CRLF or MaxLen bytes.  CR or CRLF are not stored in
	Buffer.
*/
int netbox_GetLine(PNETBOXPROFILE NetProfile, char *Buffer, int MaxLen)
{
        int Index=0, len, Continue=1;
        do{
                if((len=netbox_iGetBuffer(NetProfile, Buffer+Index,1))<=0) {
			return len;
                } else {
			switch(Buffer[Index]) {
				case 0x0a:
					/* Stop reading */
                        		if(Buffer[Index-1]==0x0d)
						Buffer[Index-1]='\0';
					else
						Buffer[Index]='\0';
					Continue=0;
					break;	
                        }
			Index += len;
                }
        } while(Continue==1 && Index < MaxLen);
        return strlen(Buffer);
}

